import logging.config
import os
import sys
from argparse import _VersionAction
from concurrent.futures import ThreadPoolExecutor
from textwrap import dedent

import tornado.ioloop
import tornado.web
from flask import current_app as flask_app
from tornado import autoreload
from tornado.httpserver import HTTPServer
from tornado.wsgi import WSGIContainer

from ..autossl import AutoHTTPSServer
from ..core import workers
from ..model import QUERIES
from ..model import configure as configure_db_session
from ..toolkit import perf, proctitle, taskmanager
from ..toolkit import validators as v
from ..toolkit.app import BaseApplication, define_core_arguments
from ..toolkit.configuration import MergedConfiguration, OptionSpec
from ..toolkit.errors import UserError
from ..toolkit.signing import load_private_key
from ..toolkit.tasklist.sqlite3_engine import TaskListSQLite3Engine
from ..version import __version__, format_version, inspect_versions
from ..web.flask import finalize_app as finalize_flask_app
from ..web.tornado import Error404Handler, TemplateRenderer
from ..web.tornado import app as tornado_app

logger = logging.getLogger("temboardui")


class TemboardApplication(BaseApplication):
    PROGRAM = "temboard"
    VERSION = __version__

    DEFAULT_CONFIGFILES = [
        "/etc/temboard/temboard.conf",
        "temboard.conf",
        ".config/temboard.conf",
    ]
    DEFAULT_PLUGINS = [
        "activity",
        "dashboard",
        "monitoring",
        "pgconf",
        "maintenance",
        "statements",
    ]

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.config = TemboardUIConfiguration()

    def define_arguments(self, parser):
        define_core_arguments(parser)
        parser.add_argument(
            "-V", "--version", action=VersionAction, help="show version and exit"
        )
        # Chain up for sub-commands arguments initialization.
        super().define_arguments(parser)

    def main(self, argv, environ):
        # C O N F I G U R A T I O N

        parser = self.create_parser(
            description=dedent("""\
            temBoard UI %s.

            COMMAND is optional. Default command is `serve`, the combined
            service. See available commands below.

            """)
            % __version__
        )
        self.define_arguments(parser)

        args = parser.parse_args(argv)

        command_name = getattr(args, "command_fullname", "serve")
        command = self.commands[command_name]

        if command.is_service:
            proctitle.init("%s: " % self.PROGRAM)

        environ = map_pgvars(environ)
        self.bootstrap(args=args, environ=environ, service=command.is_service)
        self.log_versions()

        # T A S K   M A N A G E R

        task_queue = taskmanager.Queue()
        event_queue = taskmanager.Queue()

        self.worker_pool = taskmanager.WorkerPoolService(
            app=self, task_queue=task_queue, event_queue=event_queue
        )
        self.worker_pool.add(workers)
        self.services.append(self.worker_pool)
        self.scheduler = taskmanager.SchedulerService(
            app=self, task_queue=task_queue, event_queue=event_queue
        )
        self.scheduler.add(workers)
        self.services.append(self.scheduler)

        self.webservice = TornadoService(self)

        # TaskList engine setup must be done before we load the plugins
        self.scheduler.task_list_engine = TaskListSQLite3Engine(
            os.path.join(self.config.temboard["home"], "server_tasks.db")
        )

        self.apply_config()

        return command.main(args)

    def apply_config(self):
        firstrun = not hasattr(self, "tornado_app")
        if firstrun:
            # For now, just create web app once.
            self.tornado_app = bootstrap_tornado_app(tornado_app, self.config)
            self.tornado_app.executor = ThreadPoolExecutor(12)
            self.tornado_app.temboard_app = self

        super().apply_config()

        if firstrun:
            finalize_tornado_app(tornado_app, self.config)
            finalize_flask_app()  # Uses current_app thread local

        self.tornado_app.engine = configure_db_session(self.config.repository)

    def log_versions(self):
        versions = inspect_versions()
        logger.debug("Running on %s %s.", versions["distname"], versions["distversion"])
        logger.debug(
            "Using Python %s (%s), Tornado %s and Flask %s .",
            versions["python"],
            versions["pythonbin"],
            versions["tornado"],
            versions["flask"],
        )
        logger.debug(
            "Using libpq %s, Psycopg2 %s and SQLAlchemy %s .",
            versions["libpq"],
            versions["psycopg2"],
            versions["sqlalchemy"],
        )


class TemboardUIConfiguration(MergedConfiguration):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._signing_key = None

    @property
    def signing_key(self):
        # Lazy load signing key.
        if not self._signing_key:
            self.load_signing_key()
        return self._signing_key

    def load_signing_key(self):
        path = self.temboard.signing_private_key
        logger.debug("Loading signing key from %s.", path)
        try:
            fo = open(path, "rb")
        except OSError as e:
            raise UserError("Failed to load signing key: %s" % e)

        with fo:
            self._signing_key = load_private_key(fo.read())


def bootstrap_tornado_app(tornado_app, config):
    tornado_app.config = config

    base_path = os.path.dirname(os.path.dirname(__file__))
    # Load handlers
    __import__("temboardui.handlers.notification")
    __import__("temboardui.handlers.settings.metadata")

    tornado_app.configure(
        cookie_secret=config.temboard["cookie_secret"],
        debug=config.logging.debug,
        template_path=base_path + "/templates",
        default_handler_class=Error404Handler,
    )

    return tornado_app


def finalize_tornado_app(tornado_app, config):
    base_path = os.path.dirname(os.path.dirname(__file__))
    handlers = [
        (
            r"/css/(.*)",
            tornado.web.StaticFileHandler,
            {"path": base_path + "/static/dist/css"},
        ),
        (
            r"/js/(.*)",
            tornado.web.StaticFileHandler,
            {"path": base_path + "/static/dist/js"},
        ),
        (
            r"/images/(.*)",
            tornado.web.StaticFileHandler,
            {"path": base_path + "/static/dist/images"},
        ),
        # Path needs a (unused) path parameter, not used by subclass
        # SingleFileHandler.
        (
            r"/(signing.key)",
            SingleFileHandler,
            {"path": config.temboard.signing_public_key},
        ),
    ]

    # Append rules *after* plugins because plugins shares same namespace for
    # static rules, i.e. /js/.* is a fallback for /js/dashboard/.*.
    tornado_app.add_rules(handlers)

    # Fallback to Flask
    tornado_app.add_rules(
        [
            (
                r"/.*",
                tornado.web.FallbackHandler,
                {"fallback": WSGIContainer(flask_app.wsgi_app)},
            )
        ]
    )

    TemplateRenderer.GLOBAL_NAMESPACE["vitejs"] = flask_app.vitejs


def map_pgvars(environ):
    pgvar_map = dict(
        PGHOST="TEMBOARD_REPOSITORY_HOST",
        PGPORT="TEMBOARD_REPOSITORY_PORT",
        PGUSER="TEMBOARD_REPOSITORY_USER",
        PGPASSWORD="TEMBOARD_REPOSITORY_PASSWORD",
        PGDATABASE="TEMBOARD_REPOSITORY_DBNAME",
    )
    mapped = environ.copy()
    for pgvar, tbvar in list(pgvar_map.items()):
        if tbvar in environ:
            continue

        try:
            mapped[tbvar] = environ[pgvar]
            logger.debug("Read %s from environ.", pgvar)
        except KeyError:
            pass

    return mapped


class TornadoService:
    name = "web"

    def __init__(self, app):
        self.app = app
        # Ref to services.BackgroundManager
        self.background = None
        self.perf = perf.PerfCounters.setup(service=self.name)

    def __str__(self):
        return self.name

    # Interface for services.run()
    def create_loop(self):
        return tornado.ioloop.IOLoop.instance()

    def setup(self, sgm, bg):
        self.background = bg
        if self.perf:
            sgm.register(self.perf)
            self.perf.run()

        flask_app.vitejs.read_manifest()

        config = self.app.config
        if config.temboard.ssl_key_file:
            ssl_ctx = {
                "certfile": config.temboard.ssl_cert_file,
                "keyfile": config.temboard.ssl_key_file,
                "ciphers": ":".join(
                    [
                        # From Mozilla SSL configuration generator. 2023-07-28
                        "ECDHE-ECDSA-AES128-GCM-SHA256",
                        "ECDHE-RSA-AES128-GCM-SHA256",
                        "ECDHE-ECDSA-AES256-GCM-SHA384",
                        "ECDHE-RSA-AES256-GCM-SHA384",
                        "ECDHE-ECDSA-CHACHA20-POLY1305",
                        "ECDHE-RSA-CHACHA20-POLY1305",
                        "DHE-RSA-AES128-GCM-SHA256",
                        "DHE-RSA-AES256-GCM-SHA384",
                        "DHE-RSA-CHACHA20-POLY1305",
                    ]
                ),
            }
            server = AutoHTTPSServer(self.app.tornado_app, ssl_options=ssl_ctx)
        else:
            # Use plain HTTP for development to mix vitejs dev server and
            # tornado server in the same browser page.
            server = HTTPServer(self.app.tornado_app)
        try:
            server.listen(config.temboard.port, address=config.temboard.address)
        except OSError as e:
            logger.error("FATAL: " + str(e) + ". Quit")
            sys.exit(3)

        # Automatically reload modified modules (from Tornado's
        # Application.__init__). This code must be done here *after*
        # daemonize, because it instanciates ioloop for current PID.
        if self.app.tornado_app.settings.get("autoreload"):
            self._setup_autoreload()
            logger.info("Enabling Tornado's autoreload.")
            autoreload.start()

        logger.info(
            "Serving temboardui on http%s://%s:%d",
            "s" if self.app.config.temboard.ssl_cert_file else "",
            self.app.config.temboard.address,
            self.app.config.temboard.port,
        )

    def _setup_autoreload(self):
        autoreload.add_reload_hook(self._autoreload_hook)

        autoreload.watch(self.app.config.temboard.configfile)
        autoreload.watch(self.app.config.temboard.signing_public_key)
        autoreload.watch(self.app.config.temboard.signing_private_key)
        autoreload.watch(flask_app.vitejs.manifest_path)

        for path in self._iter_template_files():
            autoreload.watch(path)

        for path in QUERIES.iter_files():
            autoreload.watch(path)

    def _autoreload_hook(self):
        if not self.background:
            return
        logger.debug("Stopping background service before reloading.")
        self.background.stop()

    def _iter_template_files(self):
        rootpkg = __import__(__name__)
        rootdir = rootpkg.__path__[0]
        for dirpath, dirnames, filenames in os.walk(rootdir):
            for filename in filenames:
                if filename.endswith(".html"):
                    yield dirpath + "/" + filename

    # Interface for SignalMultiplexer
    def sighup_handler(self, *a):
        self.app.reload()


class SingleFileHandler(tornado.web.StaticFileHandler):
    @classmethod
    def get_absolute_path(cls, root, *a):
        return root

    def validate_absolute_path(self, root, absolute_path):
        if not os.path.exists(absolute_path):
            logger.warning("Inexistant file %s", absolute_path)
            raise tornado.web.HTTPError(404)
        return absolute_path


class VersionAction(_VersionAction):
    def __call__(self, parser, *_):
        print(format_version().strip())
        parser.exit()


def cookie_secret(raw):
    length = len(raw)
    if length < 10:
        raise ValueError("cookie secret is shorter than 10 chars.")
    if length > 128:
        raise ValueError("cookie secret is longer than 128 chars.")
    return raw


def list_options_specs():
    s = "temboard"
    # Manage plugin list here until we use plugin entrypoint.
    yield OptionSpec(
        s, "plugins", default=TemboardApplication.DEFAULT_PLUGINS, validator=v.jsonlist
    )
    yield OptionSpec(s, "address", default="0.0.0.0", validator=v.address)
    yield OptionSpec(s, "port", validator=v.port, default=8888)
    yield OptionSpec(s, "ssl_cert_file", default=None, validator=v.file_)
    yield OptionSpec(s, "ssl_key_file", default=None, validator=v.file_)
    yield OptionSpec(s, "ssl_ca_cert_file", validator=v.file_)
    yield OptionSpec(
        s, "signing_private_key", default="signing-private.pem", validator=v.path
    )
    yield OptionSpec(
        s, "signing_public_key", default="signing-public.pem", validator=v.path
    )
    yield OptionSpec(s, "cookie_secret", validator=cookie_secret)
    home = os.environ.get("HOME", "/var/lib/temboard")
    yield OptionSpec(s, "home", default=home, validator=v.writeabledir)

    s = "auth"
    yield OptionSpec(s, "allowed_ip", default="127.0.0.0/8", validator=v.commalist)

    s = "repository"
    yield OptionSpec(s, "host", default="/var/run/postgresql")
    yield OptionSpec(s, "instance", default="main")
    yield OptionSpec(s, "port", default=5432, validator=v.port)
    yield OptionSpec(s, "user", default="temboard")
    yield OptionSpec(s, "password", default="temboard")
    yield OptionSpec(s, "dbname", default="temboard")

    s = "notifications"
    yield OptionSpec(s, "smtp_host", default=None)
    yield OptionSpec(s, "smtp_port", default=None, validator=v.port)
    yield OptionSpec(s, "smtp_tls", default=False, validator=v.boolean)
    yield OptionSpec(s, "smtp_login", default=None)
    yield OptionSpec(s, "smtp_password", default=None)
    yield OptionSpec(s, "smtp_from_addr", default=None)
    yield OptionSpec(s, "twilio_account_sid", default=None)
    yield OptionSpec(s, "twilio_auth_token", default=None)
    yield OptionSpec(s, "twilio_from", default=None)

    s = "monitoring"
    yield OptionSpec(s, "purge_after", default=730, validator=v.nday)

    s = "statements"
    yield OptionSpec(s, "purge_after", default=7, validator=v.nday)


app = TemboardApplication(specs=list_options_specs())



import logging
from time import sleep

import flask
from flask import current_app as app
from flask import g, jsonify, make_response, redirect, render_template, request
from tornado.web import create_signed_value

from temboardui.application import gen_cookie, get_role_by_auth, hash_password
from temboardui.errors import TemboardUIError

from ...model import orm
from ...toolkit import validators
from ..flask import admin_required, anonymous_allowed, transaction, validating

logger = logging.getLogger(__name__)


@app.route("/logout")
def logout():
    response = make_response(redirect("/"))
    response.delete_cookie("temboard")
    return response


@app.route("/login")
@anonymous_allowed
def login():
    if g.current_user:
        return redirect("/home")
    return render_template("login.html", headerbar=False)


@app.route(r"/json/login", methods=["POST"])
@anonymous_allowed
def json_login():
    username = request.json["username"]
    password = request.json["password"]

    response = make_response(jsonify({"message": "OK"}))
    passhash = hash_password(username, password).decode("utf-8")

    try:
        role = get_role_by_auth(g.db_session, username, passhash)
    except TemboardUIError as e:
        logger.error("Login failed: %s", e)
        response = make_response(jsonify({"error": "Wrong username/password."}))
        response.status_code = 401
        # Mitigate dictionnaries attacks.
        sleep(1)
        return response

    logger.info("Role '%s' authentificated.", role.role_name)
    secret_cookie = create_signed_value(
        app.temboard.config.temboard.cookie_secret,
        "temboard",
        gen_cookie(role.role_name, passhash),
    )
    response.set_cookie("temboard", secret_cookie.decode(), secure=True)
    return response


@app.route("/json/users")
@admin_required
def get_users():
    return jsonify([u.asdict() for u in orm.Role.all().with_session(g.db_session)])


@app.route("/json/users", methods=["POST"])
@admin_required
@transaction
def post_user():
    if "password" not in request.json:
        raise flask.abort(400, "Password required.")
    role = orm.Role(groups=[])
    g.db_session.add(role)
    return put_user(user=role)


@app.route("/json/users/<name>")
@admin_required
def get_user(name):
    user = orm.Role.get(name).with_session(g.db_session).one_or_none()
    if user is None:
        flask.abort(404, "No such user.")
    return flask.jsonify(user.asdict())


@app.route("/json/users/<name>", methods=["PUT"])
@admin_required
@transaction
def put_user(name=None, user=None):
    if user is None:
        user = orm.Role.get(name).with_session(g.db_session).one_or_none()
    if user is None:
        flask.abort(404, "No such user.")

    j = request.json
    user.is_admin = j["is_admin"]
    user.is_active = j["is_active"]
    if j["name"] in {"temboard"}:
        raise flask.abort(400, "Reserved user name.")

    with validating():
        user.role_name = validators.slug(j["name"])
        if j["email"]:
            user.role_email = validators.email(j["email"])
        elif user.role_email:
            user.role_email = None  # Remove email
        if j["phone"]:
            user.role_phone = validators.phone(j["phone"])
        if j.get("password"):
            validators.password(j["password"])
            if j["password"] != j.get("password2"):
                raise ValueError("password mismatch")
            user.role_password = hash_password(user.role_name, j["password"]).decode(
                "utf-8"
            )
    g.db_session.flush()

    return flask.jsonify(user.asdict())


@app.route("/json/users/<name>", methods=["DELETE"])
@admin_required
@transaction
def delete_user(name):
    result = g.db_session.execute(orm.Role.delete(name))
    if result.rowcount == 0:
        flask.abort(404, "No such user.")
    return flask.jsonify()


@app.route("/json/groups/<path:groupname>/members/<username>")
@admin_required
def get_group_membership(groupname, username):
    row = g.db_session.execute(orm.Group.select_membership(groupname, username)).first()
    if not row:
        flask.abort(404, "No such membership.")

    return flask.jsonify({k: getattr(row, k) for k in row.keys()})


@app.route("/json/groups/<path:groupname>/members", methods=["POST"])
@admin_required
@transaction
def post_group_membership(groupname):
    gr = orm.Group.get(groupname).with_session(g.db_session).one_or_none()
    if not gr:
        flask.abort(404, "No such group.")

    g.db_session.execute(gr.insert_member(request.json["username"]))
    return flask.jsonify(
        username=request.json["username"], groupname=gr.name, profile=gr.description
    )


@app.route("/json/groups/<path:groupname>/members/<username>", methods=["DELETE"])
@admin_required
@transaction
def delete_group_membership(groupname, username):
    """Remove a user from a group."""
    result = g.db_session.execute(orm.Group.delete_member(groupname, username))
    if not result.rowcount:
        flask.abort(404, "No such membership.")

    return flask.jsonify()



import logging

from .agentclient import TemboardAgentClient
from .model import Session, worker_engine
from .model.orm import Instance
from .toolkit.taskmanager import WorkerSet
from .toolkit.utils import utcnow

logger = logging.getLogger(__name__)
workers = WorkerSet()


@workers.register(pool_size=1)
def refresh_discover(app, address, port):
    session = Session(bind=worker_engine(app.config.repository))
    instance = Instance.get(address, port).with_session(session).one()
    client = TemboardAgentClient.factory(app.config, address, port)
    try:
        logger.info("Discovering %s.", instance)
        response = client.get("/discover")
        response.raise_for_status()
    except (OSError, ConnectionError, client.Error) as e:
        logger.error("Failed to discover %s: %s", instance, e)
        logger.error("Agent or host may be down or misconfigured.")
        return

    data = response.json()

    discover_etag = response.headers.get("ETag")
    if discover_etag == instance.discover_etag:
        logger.info("Discover data up to date for %s.", instance)
        if session.is_modified(instance):
            session.commit()
        return

    instance.discover = data
    instance.discover_etag = discover_etag
    instance.discover_date = utcnow()
    session.commit()

    logger.info("Updated discover data for %s.", instance)



# Flask routes
from flask import current_app

from ...web.flask import apikey_allowed, instance_proxy


@instance_proxy.route("/monitoring/metrics")
@apikey_allowed
def get_metrics():
    return current_app.instance.proxy()



import functools
import logging
from textwrap import dedent

logger = logging.getLogger(__name__.replace(".db", ""))


def handle_keyerror(fun):
    metric_name = fun.__name__.replace("insert_metric_", "")

    @functools.wraps(fun)
    def wrapper(*a, **kw):
        try:
            return fun(*a, **kw)
        except KeyError as e:
            logger.warning("Missing point %s for %s.", e.args[0], metric_name)

    return wrapper


def insert_availability(session, dt, instance_id, available):
    session.execute(
        dedent("""
        SELECT monitoring.insert_instance_availability(
            :dt,
            :instance_id,
            :available
        )
        """),
        dict(dt=dt, instance_id=instance_id, available=available),
    )


@handle_keyerror
def insert_metric_sessions(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_sessions_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            dbname=metric["dbname"],
            row=(
                None,
                metric["active"],
                metric["waiting"],
                metric["idle"],
                metric["idle_in_xact"],
                metric["idle_in_xact_aborted"],
                metric["fastpath"],
                metric["disabled"],
                metric["no_priv"],
            ),
        ),
    )


@handle_keyerror
def insert_metric_xacts(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_xacts_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            dbname=metric["dbname"],
            row=(
                None,
                str(metric["measure_interval"]),
                metric["n_commit"],
                metric["n_rollback"],
            ),
        ),
    )


@handle_keyerror
def insert_metric_locks(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_locks_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            dbname=metric["dbname"],
            row=(
                None,
                metric["access_share"],
                metric["row_share"],
                metric["row_exclusive"],
                metric["share_update_exclusive"],
                metric["share"],
                metric["share_row_exclusive"],
                metric["exclusive"],
                metric["access_exclusive"],
                metric["siread"],
                metric["waiting_access_share"],
                metric["waiting_row_share"],
                metric["waiting_row_exclusive"],
                metric["waiting_share_update_exclusive"],
                metric["waiting_share"],
                metric["waiting_share_row_exclusive"],
                metric["waiting_exclusive"],
                metric["waiting_access_exclusive"],
            ),
        ),
    )


@handle_keyerror
def insert_metric_blocks(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_blocks_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            dbname=metric["dbname"],
            row=(
                None,
                str(metric["measure_interval"]),
                metric["blks_read"],
                metric["blks_hit"],
                metric["hitmiss_ratio"],
            ),
        ),
    )


@handle_keyerror
def insert_metric_bgwriter(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_bgwriter_current
            VALUES (:dt, :instance_id, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            row=(
                None,
                str(metric["measure_interval"]),
                metric["checkpoints_timed"],
                metric["checkpoints_req"],
                metric["checkpoint_write_time"],
                metric["checkpoint_sync_time"],
                metric["buffers_checkpoint"],
                metric["buffers_clean"],
                metric["maxwritten_clean"],
                metric["buffers_backend"],
                metric["buffers_backend_fsync"],
                metric["buffers_alloc"],
                metric["stats_reset"],
            ),
        ),
    )


@handle_keyerror
def insert_metric_db_size(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_db_size_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            dbname=metric["dbname"],
            row=(None, metric["size"]),
        ),
    )


@handle_keyerror
def insert_metric_tblspc_size(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_tblspc_size_current
            VALUES (:dt, :instance_id, :spcname, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            spcname=metric["spcname"],
            row=(None, metric["size"]),
        ),
    )


@handle_keyerror
def insert_metric_fs_size(session, host_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_filesystems_size_current
            VALUES (:dt, :host_id, :mount_point, :row)
        """),
        dict(
            dt=metric["datetime"],
            host_id=host_id,
            mount_point=metric["mount_point"],
            row=(None, metric["used"], metric["total"], metric["device"]),
        ),
    )


@handle_keyerror
def insert_metric_temp_files_size_delta(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_temp_files_size_delta_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            dbname=metric["dbname"],
            row=(None, str(metric["measure_interval"]), metric["size"]),
        ),
    )


@handle_keyerror
def insert_metric_wal_files(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_wal_files_current
            VALUES (:dt, :instance_id, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            row=(
                None,
                str(metric["measure_interval"]),
                metric["written_size"],
                metric["current_location"],
                metric["total"],
                metric["archive_ready"],
                metric["total_size"],
            ),
        ),
    )


@handle_keyerror
def insert_metric_cpu(session, host_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_cpu_current
            VALUES (:dt, :host_id, :cpu, :row)
        """),
        dict(
            dt=metric["datetime"],
            host_id=host_id,
            cpu=metric["cpu"],
            row=(
                None,
                str(metric["measure_interval"]),
                metric["time_user"],
                metric["time_system"],
                metric["time_idle"],
                metric["time_iowait"],
                metric["time_steal"],
            ),
        ),
    )


@handle_keyerror
def insert_metric_process(session, host_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_process_current
            VALUES (:dt, :host_id, :row)
        """),
        dict(
            dt=metric["datetime"],
            host_id=host_id,
            row=(
                None,
                str(metric["measure_interval"]),
                metric["context_switches"],
                metric["forks"],
                metric["procs_running"],
                metric["procs_blocked"],
                metric["procs_total"],
            ),
        ),
    )


@handle_keyerror
def insert_metric_memory(session, host_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_memory_current
            VALUES (:dt, :host_id, :row)
        """),
        dict(
            dt=metric["datetime"],
            host_id=host_id,
            row=(
                None,
                metric["mem_total"],
                metric["mem_used"],
                metric["mem_free"],
                metric["mem_buffers"],
                metric["mem_cached"],
                metric["swap_total"],
                metric["swap_used"],
            ),
        ),
    )


@handle_keyerror
def insert_metric_loadavg(session, host_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_loadavg_current
            VALUES (:dt, :host_id, :row)
        """),
        dict(
            dt=metric["datetime"],
            host_id=host_id,
            row=(None, metric["load1"], metric["load5"], metric["load15"]),
        ),
    )


@handle_keyerror
def insert_metric_vacuum_analyze(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_vacuum_analyze_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            dbname=metric["dbname"],
            row=(
                None,
                str(metric["measure_interval"]),
                metric["n_vacuum"],
                metric["n_analyze"],
                metric["n_autovacuum"],
                metric["n_autoanalyze"],
            ),
        ),
    )


@handle_keyerror
def insert_metric_replication_lag(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_replication_lag_current
            VALUES (:dt, :instance_id, :row)
        """),
        dict(dt=metric["datetime"], instance_id=instance_id, row=(None, metric["lag"])),
    )


@handle_keyerror
def insert_metric_replication_connection(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_replication_connection_current
            VALUES (:dt, :instance_id, :upstream, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            upstream=metric["upstream"],
            row=(None, metric["connected"]),
        ),
    )


@handle_keyerror
def insert_metric_heap_bloat(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_heap_bloat_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            dbname=metric["dbname"],
            row=(None, metric["ratio"]),
        ),
    )


@handle_keyerror
def insert_metric_btree_bloat(session, instance_id, metric):
    session.execute(
        dedent("""
            INSERT INTO monitoring.metric_btree_bloat_current
            VALUES (:dt, :instance_id, :dbname, :row)
        """),
        dict(
            dt=metric["datetime"],
            instance_id=instance_id,
            dbname=metric["dbname"],
            row=(None, metric["ratio"]),
        ),
    )


def get_host_id(session, hostname):
    row = session.execute(
        "SELECT host_id FROM monitoring.hosts WHERE hostname = :hostname",
        dict(hostname=hostname),
    ).fetchone()
    return row[0] if row else None


def get_instance_id(session, host_id, port):
    row = session.execute(
        dedent("""
            SELECT instance_id
            FROM monitoring.instances
            WHERE host_id = :host_id AND port = :port
        """),
        dict(host_id=host_id, port=port),
    ).fetchone()
    return row[0] if row else None


def append_state_changes(session, dt, check_id, state, key, value, warning, critical):
    session.execute(
        dedent("""
            SELECT monitoring.append_state_changes(
                :datetime, :check_id, :state, :key, :value, :warning, :critical
            )
        """),
        dict(
            datetime=dt,
            check_id=check_id,
            state=state,
            key=key,
            value=value,
            warning=warning,
            critical=critical,
        ),
    )


def purge_check_states(session, check_id, keys):
    session.execute(
        dedent("""
            DELETE FROM monitoring.check_states
            WHERE check_id = :check_id AND NOT (key = ANY(:keys))
        """),
        dict(check_id=check_id, keys=keys),
    )


def undef_check_states(session, all_ids, to_keep_ids):
    session.execute(
        dedent("""
            UPDATE monitoring.check_states
            SET state = 'UNDEF'
            WHERE
                check_id = ANY(:all_ids)
                AND NOT (check_id = ANY(:to_keep_ids))
        """),
        dict(all_ids=all_ids, to_keep_ids=to_keep_ids),
    )



import logging
from itertools import chain

from flask import current_app

from ..toolkit.app import SubCommand
from .app import app

logger = logging.getLogger(__name__)


@app.command
class Routes(SubCommand):
    """List HTTP routes map."""

    def define_arguments(self, parser):
        parser.add_argument(
            "--sort",
            action="store_true",
            default=False,
            help="Sort routes alphabetically",
        )

    def main(self, args):
        rules = self.app.tornado_app.wildcard_router.rules
        routes = chain(iter_tornado_routes(rules), iter_flask_routes())

        if args.sort:
            logger.debug("Sorting routes alphabetically.")
            routes = sorted(routes, key=lambda x: (x[1], x[0]))
        else:
            logger.debug("Listing routes by matching order.")

        for method, path in routes:
            print("    %6.6s %-64s" % (method, path))

        return 0


def iter_flask_routes():
    for rule in current_app.url_map.iter_rules():
        for method in rule.methods:
            if method in ("OPTIONS", "HEAD"):
                continue
            yield method, rule.rule


def iter_tornado_routes(rules):
    for rule in rules:
        if "fallback" in rule.target_kwargs:
            # Skip fallback to Flask routes.
            continue

        # Fallback for static handlers
        methods = rule.target_kwargs.get("methods", ["GET"])
        path = rule.matcher.regex.pattern
        for method in methods:
            yield method, path



import logging

from temboardui.web.tornado import HTTPError, csvify

from ..chartdata import get_metric_data_csv, get_unavailability_csv
from ..tools import get_request_ids, parse_start_end
from . import blueprint, render_template

logger = logging.getLogger(__name__)


@blueprint.instance_route("/monitoring")
def index(request):
    request.instance.check_active_plugin("monitoring")
    request.instance.fetch_status()
    return render_template(
        "index.html",
        role=request.current_user,
        instance=request.instance,
        plugin="monitoring",
    )


@blueprint.instance_route("/monitoring/unavailability")
def unavailability(request):
    try:
        host_id, instance_id = get_request_ids(request)
    except NameError as e:
        logger.info("%s. No data.", e)
        return csvify(data=[])

    start, end = parse_start_end(request)
    data = get_unavailability_csv(request.db_session, start, end, host_id, instance_id)
    return csvify(data)


@blueprint.instance_route(r"/monitoring/data/([a-z\-_.0-9]{1,64})$")
def data_metric(request, metric_name):
    key = request.handler.get_argument("key", default=None)
    try:
        host_id, instance_id = get_request_ids(request)
    except NameError as e:
        logger.info("%s. No data.", e)
        return csvify(data=[])

    start, end = parse_start_end(request)
    try:
        data = get_metric_data_csv(
            request.db_session,
            metric_name,
            start,
            end,
            host_id=host_id,
            instance_id=instance_id,
            key=key,
        )
    except IndexError:
        raise HTTPError(404, "Unknown metric.")

    return csvify(data=data)



import logging

from flask import current_app as app
from flask import g, render_template

from ..flask import instance_routes

logger = logging.getLogger(__name__)


@instance_routes.route("/about")
def instance_about():
    app.instance.fetch_status()
    return render_template(
        "instance-about.html",
        instance_name=g.instance.__str__(),
        pg_data=g.instance.pg_data,
        pg_version_summary=g.instance.pg_version_summary,
        discover=g.instance.discover,
        environment=g.instance.environment.name,
    )



import os

from setuptools import __version__ as setuptoolsv
from setuptools import find_packages, setup

# Load version number
__version__ = None
setup_path = os.path.dirname(os.path.realpath(__file__))
exec(open(os.path.join(setup_path, "temboardui", "version.py")).read())

if setuptoolsv < "1.0":
    __version__ = __version__.replace("+", ".")

install_requires = [
    "cryptography",
    "flask",
    "python-dateutil>=1.5",
    "setuptools",
    # There is no hard dependency on psycopg2 to allow using
    # psycopg2-binary instead. psycopg2 is not provided by psycopg2-binary
    # and there is no way to state an OR dependency in Python. It's up to
    # the user or package manager to ensure psycopg2 dependency. See
    # documentation.
    "sqlalchemy>=1.3.2,<2",
    # _verify_cert in SSLIOStream is removed on Tornado 6.4+ on Python 3+.
    "tornado>=6.0.2,<6.4",
    "future",
]

SETUP_KWARGS = dict(
    name="temboard",
    version=__version__,  # noqa, imported by execfile.
    description="temBoard User Interface.",
    author="Dalibo",
    author_email="contact@dalibo.com",
    license="PostgreSQL",
    url="https://labs.dalibo.com/temboard",
    classifiers=[
        "Intended Audience :: System Administrators",
        "License :: OSI Approved",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.6",
        "Topic :: Database :: Database Engines/Servers",
        "Topic :: System :: Monitoring",
    ],
    install_requires=install_requires,
    include_package_data=True,
    zip_safe=False,
    data_files=[
        (
            "share/temboard",
            [
                "share/auto_configure.sh",
                "share/create_repository.sh",
                "share/postinst.sh",
                "share/preun.sh",
                "share/purge.sh",
            ],
        ),
        (
            "share/temboard/sql/",
            [
                "share/sql/dev-fixture.sql",
                "share/sql/upgrade-monitoring-purge-instances.sql",
                "share/sql/reassign.sql",
            ],
        ),
        (
            "share/temboard/quickstart/",
            [
                "share/temboard_CHANGEME.key",
                "share/temboard_CHANGEME.pem",
                "share/temboard_ca_certs_CHANGEME.pem",
                "share/temboard.conf",
            ],
        ),
        ("lib/systemd/system", ["packaging/temboard.service"]),
    ],
    entry_points={
        "console_scripts": ["temboard = temboardui.__main__:main"],
        "temboardui.plugins": [
            "activity = temboardui.plugins.activity:ActivityPlugin",
            "dashboard = temboardui.plugins.dashboard:DashboardPlugin",
            "maintenance = temboardui.plugins.maintenance:MaintenancePlugin",
            "monitoring = temboardui.plugins.monitoring:MonitoringPlugin",
            "pgconf = temboardui.plugins.pgconf:PGConfPlugin",
            "statements = temboardui.plugins.statements:StatementsPlugin",
        ],
    },
)


if __name__ == "__main__":
    setup(
        long_description=open("README.md", encoding="utf-8").read(),
        long_description_content_type="text/markdown",
        packages=find_packages(),
        **SETUP_KWARGS,
    )



import sys
from platform import python_version

__version__ = "9.0rc1"


# This output is parsed by tests/conftest.py::pytest_report_header.
VERSION_FMT = """\
temBoard %(temboard)s (%(temboardbin)s)
System %(distname)s %(distversion)s
Python %(python)s (%(pythonbin)s)
cryptography %(cryptography)s
Tornado %(tornado)s
Flask %(flask)s
libpq %(libpq)s
psycopg2 %(psycopg2)s
SQLAlchemy %(sqlalchemy)s
"""


def format_version():
    return VERSION_FMT % inspect_versions()


def inspect_versions():
    import cryptography
    import flask
    import psycopg2
    import sqlalchemy
    import tornado

    from .toolkit.versions import format_pq_version, read_distinfo, read_libpq_version

    distinfos = read_distinfo()

    return dict(
        cryptography=cryptography.__version__,
        distname=distinfos["NAME"],
        distversion=distinfos.get("VERSION", "n/a"),
        flask=flask.__version__,
        libpq=format_pq_version(read_libpq_version()),
        psycopg2=psycopg2.__version__,
        python=python_version(),
        pythonbin=sys.executable,
        sqlalchemy=sqlalchemy.__version__,
        temboard=__version__,
        temboardbin=sys.argv[0],
        tornado=tornado.version,
    )



class CLIError(Exception):
    """CLI errors"""

    def __init__(self, message):
        Exception.__init__(self, message)
        self.message = str(message)


class TemboardUIError(Exception):
    def __init__(self, code, message):
        Exception.__init__(self, message)
        self.code = int(code)
        self.message = str(message)



import logging
import sys
from textwrap import dedent

from ..model import Session, orm
from ..toolkit.app import SubCommand
from ..toolkit.errors import UserError
from .app import app

logger = logging.getLogger(__name__)


@app.command
class ApiKey(SubCommand):
    """Manage API keys."""

    def main(self, args):
        raise UserError("Missing sub-command. See --help for details.")


@ApiKey.command
class Create(SubCommand):
    """Create a new API key."""

    def define_arguments(self, parser):
        parser.add_argument("--comment", help="Define a custom comment.")

    def main(self, args):
        session = Session()
        key = (
            orm.ApiKey.insert(secret=orm.ApiKey.generate_secret(), comment=args.comment)
            .with_session(session)
            .scalar()
        )
        session.commit()

        sys.stdout.write(
            dedent("""\
        Id,Secret,Comment,Expiration
        {key.id},"{key.secret}","{comment}",{edate}
        """).format(comment=key.comment or "", edate=key.edate.isoformat(), key=key)
        )


@ApiKey.command
class List(SubCommand):
    """List active keys."""

    def main(self, args):
        session = Session()

        sys.stdout.write("Id,Secret,Comment,Creation,Expiration\n")
        for key in orm.ApiKey.select_active().with_session(session).all():
            sys.stdout.write(
                dedent("""\
            {key.id},"{key.secret}","{comment}",{cdate},{edate}
            """).format(
                    comment=key.comment or "",
                    cdate=key.cdate.isoformat(),
                    edate=key.edate.isoformat(),
                    key=key,
                )
            )


@ApiKey.command
class Delete(SubCommand):
    """Delete a key."""

    def define_arguments(self, parser):
        parser.add_argument("id", type=int, help="Target key id.")

    def main(self, args):
        session = Session()
        key = orm.ApiKey.delete(args.id).with_session(session).scalar()
        session.commit()

        if key:
            logger.info("Deleted API key #%s (%s).", key.id, key.comment)
        else:
            logger.info("No API key deleted.")


@ApiKey.command
class Purge(SubCommand):
    """Purge expired."""

    def main(self, args):
        session = Session()
        res = orm.ApiKey.purge().with_session(session).all()
        count = 0
        for key in res:
            logger.info("Deleted API key #%s (%s).", key.id, key.comment)
            count += 1
        session.commit()

        if count:
            logger.info("Purged %d keys.", count)
        else:
            logger.info("No expired keys to purge.")



import logging

from flask import current_app, render_template

from ...agentclient import TemboardAgentClient
from ...web.flask import instance_routes

logger = logging.getLogger(__name__)


@instance_routes.route("/dashboard")
def dashboard():
    current_app.instance.check_active_plugin("dashboard")

    try:
        config = current_app.instance.request("/dashboard/config").json()
    except TemboardAgentClient.Error as e:
        if 404 != e.code:
            raise
        logger.debug("Fallback dashboard config.")
        config = {"history_length": 150, "scheduler_interval": 2}

    history = current_app.instance.request("/dashboard/history").json()

    current_app.instance.fetch_status()
    dashboard = history[-1] if history else {}
    return render_template(
        "dashboard.html",
        plugin="dashboard",
        config=config,
        dashboard=dashboard,
        history=history or "",
    )



import json
import logging
from decimal import Decimal
from os import path

import tornado.web
from past.utils import old_div
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.sql import case, column, extract, func, select, text

from temboardui.agentclient import TemboardAgentClient
from temboardui.model import worker_engine
from temboardui.model.orm import Instance
from temboardui.plugins.monitoring.tools import parse_start_end
from temboardui.toolkit import taskmanager
from temboardui.web.tornado import Blueprint, TemplateRenderer, jsonify

logger = logging.getLogger(__name__)
workers = taskmanager.WorkerSet()

blueprint = Blueprint()
plugin_path = path.dirname(path.realpath(__file__))
render_template = TemplateRenderer(plugin_path + "/templates")


class StatementsPlugin:
    def __init__(self, app):
        self.app = app

    def load(self):
        self.app.tornado_app.add_rules(blueprint.rules)
        self.app.tornado_app.add_rules(
            [
                (
                    r"/js/statements/(.*)",
                    tornado.web.StaticFileHandler,
                    {"path": plugin_path + "/static/js"},
                )
            ]
        )
        self.app.worker_pool.add(workers)
        self.app.scheduler.add(workers)


METAS_QUERY = text("""
    SELECT *
    FROM statements.metas
    WHERE agent_address = :agent_address
    AND agent_port = :agent_port
""")


BASE_QUERY_STATDATA = text("""
    (
        SELECT dbid, datname, (record).*
        FROM (
          SELECT psh.dbid, psh.datname, unnest(records) AS record
          FROM statements.statements_history_db psh
          WHERE coalesce_range && tstzrange(:start, :end,'[]')
          AND    agent_address = :agent_address
          AND    agent_port = :agent_port
        ) AS unnested
        WHERE tstzrange((record).ts, (record).ts, '[]')
              <@ tstzrange(:start, :end, '[]')

        UNION ALL

        SELECT dbid, datname, (record).*
        FROM statements.statements_history_current_db
        WHERE agent_address = :agent_address
        AND agent_port = :agent_port
        AND tstzrange((record).ts, (record).ts, '[]')
          <@ tstzrange(:start, :end, '[]')
    ) h
""")


@blueprint.instance_route(r"/statements/data", json=True)
def json_data_instance(request):
    start, end = parse_start_end(request)

    base_query = BASE_QUERY_STATDATA
    diffs = get_diffs_forstatdata()
    query = (
        select([column("datname"), column("dbid")] + diffs)
        .select_from(base_query)
        .group_by(column("dbid"), column("datname"))
        .having(func.max(column("calls")) - func.min(column("calls")) > 0)
    )

    statements = request.db_session.execute(
        query,
        dict(
            agent_address=request.instance.agent_address,
            agent_port=request.instance.agent_port,
            start=start,
            end=end,
        ),
    ).fetchall()
    statements = [dict(statement) for statement in statements]

    metas = request.db_session.execute(
        METAS_QUERY,
        dict(
            agent_address=request.instance.agent_address,
            agent_port=request.instance.agent_port,
        ),
    ).fetchone()
    metas = dict(metas) if metas is not None else None
    return jsonify(dict(data=statements, metas=metas))


BASE_QUERY_STATDATA_DATABASE = """
WITH first_occurence AS (
  -- first occurence for each statement
  SELECT
    distinct on (queryid, userid)
    queryid,
    userid,
    ts,
    calls,
    total_exec_time,
    shared_blks_read,
    shared_blks_hit,
    shared_blks_dirtied,
    shared_blks_written,
    local_blks_read,
    local_blks_hit,
    local_blks_dirtied,
    local_blks_written,
    temp_blks_read,
    temp_blks_written,
    blk_read_time,
    blk_write_time
  FROM
    (
      (
        SELECT
          distinct on (queryid, userid)
          queryid,
          userid,
          (record).*
        FROM
          (
            SELECT
              distinct on (queryid, userid)
              queryid,
              userid,
              unnest(records) AS record
            FROM
              statements.statements_history psh
            WHERE
              coalesce_range && tstzrange(
                (:start) :: timestamptz,
                (:end) :: timestamptz,
                '[]'
              )
              AND agent_address = :agent_address
              AND agent_port = :agent_port
              AND dbid = :dbid
              {queryidfilter}
            order by
              queryid,
              userid,
              coalesce_range
          ) AS unnested
        WHERE
          tstzrange(
            (record).ts,
            (record).ts,
            '[]'
          ) <@ tstzrange(
            (:start) :: timestamptz,
            (:end) :: timestamptz,
            '[]'
          )
        order by
          queryid,
          userid,
          ts
      )
      UNION ALL
        (
          SELECT
            distinct on (queryid, userid)
            queryid,
            userid,
            (record).*
          FROM
            statements.statements_history_current
          WHERE
            agent_address = :agent_address
            AND agent_port = :agent_port
            AND dbid = :dbid
            AND tstzrange(
              (record).ts,
              (record).ts,
              '[]'
            ) <@ tstzrange(
              (:start) :: timestamptz,
              (:end) :: timestamptz,
              '[]'
            )
            {queryidfilter}
          order by
            queryid,
            userid,
            ts
        )
    ) as h
  order by
    queryid,
    userid,
    ts
),
last_occurence AS (
  -- last occurence for each statement
  SELECT
    distinct on (queryid, userid)
    queryid,
    userid,
    ts,
    calls,
    total_exec_time,
    shared_blks_read,
    shared_blks_hit,
    shared_blks_dirtied,
    shared_blks_written,
    local_blks_read,
    local_blks_hit,
    local_blks_dirtied,
    local_blks_written,
    temp_blks_read,
    temp_blks_written,
    blk_read_time,
    blk_write_time
  FROM
    (
      (
        SELECT
          distinct on (queryid, userid)
          queryid,
          userid,
          (record).*
        FROM
          (
            SELECT
              distinct on (queryid, userid)
              queryid,
              userid,
              unnest(records) AS record
            FROM
              statements.statements_history psh
            WHERE
              coalesce_range && tstzrange(
                (:start) :: timestamptz,
                (:end) :: timestamptz,
                '[]'
              )
              AND agent_address = :agent_address
              AND agent_port = :agent_port
              AND dbid = :dbid
              {queryidfilter}
            order by
              queryid,
              userid,
              coalesce_range desc
          ) AS unnested
        WHERE
          tstzrange(
            (record).ts,
            (record).ts,
            '[]'
          ) <@ tstzrange(
            (:start) :: timestamptz,
            (:end) :: timestamptz,
            '[]'
          )
        order by
          queryid,
          userid,
          ts desc
      )
      UNION ALL
        (
          SELECT
            distinct on (queryid, userid)
            queryid,
            userid,
            (record).*
          FROM
            statements.statements_history_current
          WHERE
            agent_address = :agent_address
            AND agent_port = :agent_port
            AND dbid = :dbid
            AND tstzrange(
              (record).ts,
              (record).ts,
              '[]'
            ) <@ tstzrange(
              (:start) :: timestamptz,
              (:end) :: timestamptz,
              '[]'
            )
            {queryidfilter}
          order by
            queryid,
            userid,
            ts desc
        )
    ) as h
  order by
    queryid,
    userid,
    ts desc
)
SELECT
  query,
  statements.queryid::text,
  rolname,
  statements.userid::text,
  (lo.calls - fo.calls) AS calls,
  (lo.total_exec_time - fo.total_exec_time) AS total_exec_time,
  (lo.total_exec_time - fo.total_exec_time) /
    (lo.calls - fo.calls) AS mean_time,
  (
    lo.shared_blks_read - fo.shared_blks_read
  ) AS shared_blks_read,
  (
    lo.shared_blks_hit - fo.shared_blks_hit
  ) AS shared_blks_hit,
  (
    lo.shared_blks_dirtied - fo.shared_blks_dirtied
  ) AS shared_blks_dirtied,
  (
    lo.shared_blks_written - fo.shared_blks_written
  ) AS shared_blks_written,
  (
    lo.local_blks_read - fo.local_blks_read
  ) AS local_blks_read,
  (
    lo.local_blks_hit - fo.local_blks_hit
  ) AS local_blks_hit,
  (
    lo.local_blks_dirtied - fo.local_blks_dirtied
  ) AS local_blks_dirtied,
  (
    lo.local_blks_written - fo.local_blks_written
  ) AS local_blks_written,
  (
    lo.temp_blks_read - fo.temp_blks_read
  ) AS temp_blks_read,
  (
    lo.temp_blks_written - fo.temp_blks_written
  ) AS temp_blks_written,
  (
    lo.blk_read_time - fo.blk_read_time
  ) AS blk_read_time,
  (
    lo.blk_write_time - fo.blk_write_time
  ) AS blk_write_time
FROM
  first_occurence AS fo
  JOIN last_occurence AS lo
    ON fo.queryid = lo.queryid
    AND fo.userid = lo.userid
  JOIN statements.statements
    ON statements.queryid = fo.queryid
    AND statements.userid = fo.userid
  AND agent_address = :agent_address
  AND agent_port = :agent_port
  AND dbid = :dbid
WHERE (lo.calls - fo.calls) > 0;
"""


@blueprint.instance_route(r"/statements/data/([0-9]*)/([-]?[0-9]*)/([0-9]*)", json=True)
def json_data_query(request, dbid, queryid, userid):
    return json_data(request, dbid, queryid, userid)


@blueprint.instance_route(r"/statements/data/(.*)", json=True)
def json_data_database(request, dbid):
    return json_data(request, dbid)


def json_data(request, dbid, queryid=None, userid=None):
    start, end = parse_start_end(request)

    query = text("""
        SELECT DISTINCT(datname)
        FROM statements.statements
        WHERE dbid = :dbid
        AND agent_address = :agent_address
        AND agent_port = :agent_port;
    """)
    datname = request.db_session.execute(
        query,
        dict(
            agent_address=request.instance.agent_address,
            agent_port=request.instance.agent_port,
            dbid=dbid,
        ),
    ).fetchone()[0]

    params = dict(
        agent_address=request.instance.agent_address,
        agent_port=request.instance.agent_port,
        dbid=dbid,
        start=start,
        end=end,
    )

    query = BASE_QUERY_STATDATA_DATABASE
    queryidfilter = ""
    if queryid is not None and userid is not None:
        queryidfilter = "AND queryid = :queryid AND userid = :userid"
        params.update(dict(queryid=queryid, userid=userid))
    query = query.format(**dict(queryidfilter=queryidfilter))

    statements = request.db_session.execute(query, params).fetchall()
    statements = [dict(statement) for statement in statements]
    return jsonify(dict(datname=datname, data=statements))


def get_diffs_forstatdata():
    return [
        diff("calls"),
        diff("total_exec_time").label("total_exec_time"),
        (old_div(diff("total_exec_time"), diff("calls"))).label("mean_time"),
        diff("shared_blks_read"),
        diff("shared_blks_hit"),
        diff("shared_blks_dirtied"),
        diff("shared_blks_written"),
        diff("local_blks_read"),
        diff("local_blks_hit"),
        diff("local_blks_dirtied"),
        diff("local_blks_written"),
        diff("temp_blks_read"),
        diff("temp_blks_written"),
        diff("blk_read_time"),
        diff("blk_write_time"),
    ]


BASE_QUERY_STATDATA_SAMPLE_INSTANCE = text("""
    (
      SELECT *
      FROM (
        SELECT
          row_number() OVER (ORDER BY ts) AS number,
          count(*) OVER (PARTITION BY 1) AS total,
          *
        FROM (
          SELECT
            ts,
            sum(calls) AS calls,
            sum(total_exec_time) AS total_exec_time,
            sum(rows) AS rows,
            sum(shared_blks_hit) AS shared_blks_hit,
            sum(shared_blks_read) AS shared_blks_read,
            sum(shared_blks_dirtied) AS shared_blks_dirtied,
            sum(shared_blks_written) AS shared_blks_written,
            sum(local_blks_hit) AS local_blks_hit,
            sum(local_blks_read) AS local_blks_read,
            sum(local_blks_dirtied) AS local_blks_dirtied,
            sum(local_blks_written) AS local_blks_written,
            sum(temp_blks_read) AS temp_blks_read,
            sum(temp_blks_written) AS temp_blks_written,
            sum(blk_read_time) AS blk_read_time,
            sum(blk_write_time) AS blk_write_time
          FROM (
            SELECT (record).*
            FROM (
              SELECT psh.dbid, psh.coalesce_range, unnest(records) AS record
              FROM statements.statements_history_db psh
              WHERE coalesce_range && tstzrange(:start, :end,'[]')
              AND    agent_address = :agent_address
              AND    agent_port = :agent_port
            ) AS unnested
            WHERE tstzrange((record).ts, (record).ts, '[]')
                  <@ tstzrange(:start, :end, '[]')

            UNION ALL

            SELECT (record).*
            FROM statements.statements_history_current_db
            WHERE tstzrange((record).ts, (record).ts, '[]')
              <@ tstzrange(:start, :end, '[]')
            AND agent_address = :agent_address
            AND agent_port = :agent_port
          ) AS s
          GROUP BY ts
        ) AS statements_history
      ) AS sh
      WHERE number % ( int8larger((total)/(:samples +1),1) ) = 0
    ) by_instance
""")


BASE_QUERY_STATDATA_SAMPLE_DATABASE = text("""
    (
      SELECT *
      FROM (
        SELECT
          row_number() OVER (ORDER BY ts) AS number,
          count(*) OVER (PARTITION BY 1) AS total,
          *
        FROM (
          SELECT (record).*
          FROM (
            SELECT psh.dbid, psh.coalesce_range, unnest(records) AS record
            FROM statements.statements_history_db psh
            WHERE coalesce_range && tstzrange(:start, :end,'[]')
            AND    dbid = :dbid
            AND    agent_address = :agent_address
            AND    agent_port = :agent_port
          ) AS unnested
          WHERE tstzrange((record).ts, (record).ts, '[]')
                <@ tstzrange(:start, :end, '[]')

          UNION ALL

          SELECT (record).*
          FROM statements.statements_history_current_db
          WHERE tstzrange((record).ts, (record).ts, '[]')
            <@ tstzrange(:start, :end, '[]')
          AND dbid = :dbid
          AND agent_address = :agent_address
          AND agent_port = :agent_port
        ) AS statements_history
      ) AS sh
      WHERE number % ( int8larger((total)/(:samples +1),1) ) = 0
    ) by_db
""")

BASE_QUERY_STATDATA_SAMPLE_QUERY = text("""
    (
      SELECT *
      FROM (
        SELECT
          row_number() OVER (ORDER BY ts) AS number,
          count(*) OVER (PARTITION BY 1) AS total,
          *
        FROM (
          SELECT (record).*
          FROM (
            SELECT psh.dbid, psh.coalesce_range, unnest(records) AS record
            FROM statements.statements_history psh
            WHERE coalesce_range && tstzrange(:start, :end,'[]')
            AND dbid = :dbid
            AND queryid = :queryid
            AND userid = :userid
            AND agent_address = :agent_address
            AND agent_port = :agent_port
          ) AS unnested
          WHERE tstzrange((record).ts, (record).ts, '[]')
                <@ tstzrange(:start, :end, '[]')

          UNION ALL

          SELECT (record).*
          FROM statements.statements_history_current
          WHERE tstzrange((record).ts, (record).ts, '[]')
            <@ tstzrange(:start, :end, '[]')
          AND dbid = :dbid
          AND queryid = :queryid
          AND userid = :userid
          AND agent_address = :agent_address
          AND agent_port = :agent_port
        ) AS statements_history
      ) AS sh
      WHERE number % ( int8larger((total)/(:samples +1),1) ) = 0
    ) by_db
""")


def getstatdata_sample(request, mode, start, end, dbid=None, queryid=None, userid=None):
    if mode == "instance":
        base_query = BASE_QUERY_STATDATA_SAMPLE_INSTANCE

    elif mode == "db":
        base_query = BASE_QUERY_STATDATA_SAMPLE_DATABASE

    elif mode == "query":
        base_query = BASE_QUERY_STATDATA_SAMPLE_QUERY

    ts = column("ts")
    biggest = Biggest(ts)
    biggestsum = Biggestsum(ts)

    subquery = (
        select(
            [
                ts,
                biggest("ts", "0 s", "mesure_interval"),
                biggestsum("calls"),
                biggestsum("total_exec_time", label="runtime"),
                biggestsum("rows"),
                biggestsum("shared_blks_read"),
                biggestsum("shared_blks_hit"),
                biggestsum("shared_blks_dirtied"),
                biggestsum("shared_blks_written"),
                biggestsum("local_blks_read"),
                biggestsum("local_blks_hit"),
                biggestsum("local_blks_dirtied"),
                biggestsum("local_blks_written"),
                biggestsum("temp_blks_read"),
                biggestsum("temp_blks_written"),
                biggestsum("blk_read_time"),
                biggestsum("blk_write_time"),
            ]
        )
        .select_from(base_query)
        .apply_labels()
        .group_by(*([ts]))
    )

    subquery = subquery.alias()
    c = subquery.c

    greatest = func.greatest
    cols = [
        to_epoch(c.ts),
        (
            old_div(func.sum(c.calls), greatest(extract("epoch", c.mesure_interval), 1))
        ).label("calls"),
        (old_div(func.sum(c.runtime), greatest(func.sum(c.calls), 1.0))).label(
            "avg_runtime"
        ),
        (
            old_div(
                func.sum(c.runtime), greatest(extract("epoch", c.mesure_interval), 1)
            )
        ).label("load"),
        total_read(c),
        total_hit(c),
    ]

    query = (
        select(cols)
        .select_from(subquery)
        .where(c.calls != 0)
        .group_by(c.ts, c.mesure_interval)
        .order_by(c.ts)
    )

    params = dict(
        agent_address=request.instance.agent_address,
        agent_port=request.instance.agent_port,
        samples=50,
        start=start,
        end=end,
    )

    if mode == "db" or mode == "query":
        params["dbid"] = dbid
    if mode == "query":
        params["queryid"] = queryid
        params["userid"] = userid

    rows = request.db_session.execute(query, params).fetchall()
    return [dict(row) for row in rows]


def convert_decimal_to_float(data):
    # Since Postgres 14, timestamp and other data are returned as Decimal
    # rather than float. stock json does not know how to serialize Decimals.
    # For now, just preprocess data to convert decimals to float in payload
    # before serializing to json.
    if isinstance(data, Decimal):
        return float(data)
    elif isinstance(data, list):
        return [convert_decimal_to_float(v) for v in data]
    elif isinstance(data, dict):
        return {k: convert_decimal_to_float(v) for k, v in data.items()}
    else:
        return data


@blueprint.instance_route(r"/statements/chart", json=True)
def json_chart_data_instance(request):
    start, end = parse_start_end(request)

    data = getstatdata_sample(request, "instance", start, end)
    data = convert_decimal_to_float(data)
    return jsonify(dict(data=data))


@blueprint.instance_route(
    r"/statements/chart/([0-9]*)/([-]?[0-9]*)/([0-9]*)", json=True
)
def json_chart_data_query(request, dbid, queryid, userid):
    start, end = parse_start_end(request)

    data = getstatdata_sample(
        request, "query", start, end, dbid=dbid, queryid=queryid, userid=userid
    )
    data = convert_decimal_to_float(data)
    return jsonify(dict(data=data))


@blueprint.instance_route(r"/statements/chart/(.*)", json=True)
def json_chart_data_db(request, dbid):
    start, end = parse_start_end(request)

    data = getstatdata_sample(request, "db", start, end, dbid=dbid)
    data = convert_decimal_to_float(data)
    return jsonify(dict(data=data))


@blueprint.instance_route(r"/statements")
def statements(request):
    request.instance.check_active_plugin("statements")
    request.instance.fetch_status()
    return render_template(
        "index.html",
        instance=request.instance,
        plugin=__name__,
        role=request.current_user,
    )


def add_statement(session, instance, data):
    agent_id = f"{instance.agent_address}:{instance.agent_port}"
    if not data.get("data"):
        logger.info("No statement data from %s.", agent_id)
        return

    conn = session.connection().connection
    cur = conn.cursor()
    cur.execute("SET search_path TO statements")
    for statement in data.get("data"):
        query = """
            INSERT INTO statements_src_tmp
            VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,
                    %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
        """
        cur.execute(
            query,
            (
                instance.agent_address,
                instance.agent_port,
                data["snapshot_datetime"],
                statement["userid"],
                statement["rolname"],
                statement["dbid"],
                statement["datname"],
                statement["queryid"],
                statement["query"],
                statement["calls"],
                statement["total_exec_time"]
                if "total_exec_time" in statement
                else statement["total_time"],
                statement["rows"],
                statement["shared_blks_hit"],
                statement["shared_blks_read"],
                statement["shared_blks_dirtied"],
                statement["shared_blks_written"],
                statement["local_blks_hit"],
                statement["local_blks_read"],
                statement["local_blks_dirtied"],
                statement["local_blks_written"],
                statement["temp_blks_read"],
                statement["temp_blks_written"],
                # DEPRECATED: Column renamed in Postgres 17.
                statement["shared_blk_read_time"]
                if "shared_blk_read_time" in statement
                else statement["blk_read_time"],
                statement["shared_blk_write_time"]
                # DEPRECATED: Column renamed in Postgres 17.
                if "shared_blk_write_time" in statement
                else statement["blk_write_time"],
                statement["total_plan_time"]
                if "total_plan_time" in statement
                else None,
                statement["wal_records"] if "wal_records" in statement else None,
                statement["wal_fpi"] if "wal_fpi" in statement else None,
                statement["wal_bytes"] if "wal_bytes" in statement else None,
            ),
        )
    query = """SELECT process_statements(%s, %s)"""
    cur.execute(query, (instance.agent_address, instance.agent_port))
    conn.commit()


@workers.schedule(id="statements_pull_data", redo_interval=60)  # 1m
@workers.register(pool_size=1)
def pull_data_worker(app):
    engine = worker_engine(app.config.repository)
    session_factory = sessionmaker(bind=engine)
    Session = scoped_session(session_factory)
    worker_session = Session()
    instances = worker_session.query(Instance)

    if not instances:
        logger.info("No instances to pull data from.")
        return

    for instance in instances:
        plugin_names = [plugin.plugin_name for plugin in instance.plugins]

        if "statements" not in plugin_names:
            logger.debug("Skipping instance %s. Plugin disabled.", instance)
            continue

        # FIXME should be done in parallel
        try:
            pull_data_for_instance(app, worker_session, instance)
        except Exception:
            logger.exception(
                "Failed to pull data from %s:%s",
                instance.agent_address,
                instance.agent_port,
            )


@workers.register(pool_size=1)
def statements_pull1(app, host, port):
    engine = worker_engine(app.config.repository)
    session_factory = sessionmaker(bind=engine)
    Session = scoped_session(session_factory)
    worker_session = Session()
    instance = Instance.get(host, port).with_session(worker_session).first()

    try:
        pull_data_for_instance(app, worker_session, instance)
    except Exception:
        logger.exception(
            "Failed to pull data from %s:%s",
            instance.agent_address,
            instance.agent_port,
        )


def pull_data_for_instance(app, session, instance):
    agent_id = f"{instance.agent_address}:{instance.agent_port}"
    logger.info("Pulling statements from %s.", agent_id)
    client = TemboardAgentClient.factory(
        app.config, instance.agent_address, instance.agent_port
    )
    try:
        response = client.get("/statements")
        response.raise_for_status()
        add_statement(session, instance, response.json())
        logger.debug("Successfully pulled statements data for %s.", agent_id)
    except Exception as e:
        error = "Error while fetching statements from instance: "
        if hasattr(e, "read"):
            error += json.loads(e.read())["error"]
        else:
            error += str(e)

        if isinstance(e, (OSError, ConnectionError, client.Error)):
            logger.error(
                "Failed to query agent: %s. Is it running? agent=%s", error, agent_id
            )
        else:
            logger.exception("Failed to pull statements data: %s", error)

        # If statements data cannot be retrieved store the error in the
        # statements metas table
        cur = session.connection().connection.cursor()
        cur.execute("SET search_path TO statements")
        query = """
            -- Create new meta for agent if doesn't already exist
            INSERT INTO metas (agent_address, agent_port)
            VALUES (%s, %s)
            ON CONFLICT DO NOTHING;
        """
        cur.execute(query, (instance.agent_address, instance.agent_port))

        query = """
            UPDATE metas
            SET error = %s
            WHERE agent_address = %s AND agent_port = %s;
        """
        cur.execute(query, (error, instance.agent_address, instance.agent_port))
        session.connection().connection.commit()


@workers.schedule(id="statements_purge", redo_interval=24 * 60 * 60)  # 24h
@workers.register(pool_size=1)
def statements_purge_worker(app):
    """Background worker in charge of purging statements data.
    Purge policy is based on purge_after parameter from statements section.
    purge_after sets the number of days of data to keep, from now. Default is
    7 days if not set.
    """
    logger.info("Purging old data.")

    engine = worker_engine(app.config.repository)
    session_factory = sessionmaker(bind=engine)
    Session = scoped_session(session_factory)
    session = Session()

    purge_after = app.config.statements.purge_after

    # Get tablename list to purge from metric_tables_config()
    try:
        cur = session.connection().connection.cursor()
        cur.execute("SET search_path TO statements")
        cur.execute("""SELECT statements_purge(%s)""", (purge_after,))
        session.connection().connection.commit()
        logger.debug("Old statements purged successfully.")
    except Exception:
        logger.exception("Could not purge statements data:")
        raise


def to_epoch(column):
    return extract("epoch", column).label(column.name)


def diff(var):
    return (func.max(column(var)) - func.min(column(var))).label(var)


def total_measure_interval(column):
    return extract(
        "epoch",
        case([(func.min(column) == "0 second", "1 second")], else_=func.min(column)),
    )


# We use 8192 as default value for block size
# Ideally we should get this value from agent
block_size = 8192


def total_read(c):
    return (
        old_div(
            func.sum(c.shared_blks_read + c.local_blks_read + c.temp_blks_read),
            total_measure_interval(c.mesure_interval),
        )
    ).label("total_blks_read")


def total_hit(c):
    return (
        old_div(
            func.sum(c.shared_blks_hit + c.local_blks_hit),
            total_measure_interval(c.mesure_interval),
        )
    ).label("total_blks_hit")


class Biggest:
    def __init__(self, order_by):
        self.order_by = order_by

    def __call__(self, var, minval=0, label=None):
        label = label or var
        return func.greatest(
            column(var) - func.lag(column(var)).over(order_by=self.order_by), minval
        ).label(label)


class Biggestsum:
    def __init__(self, order_by):
        self.order_by = order_by

    def __call__(self, var, minval=0, label=None):
        label = label or var
        return func.greatest(
            func.sum(column(var))
            - func.lag(func.sum(column(var))).over(order_by=self.order_by),
            minval,
        ).label(label)



import logging
import os
from glob import glob
from textwrap import dedent

from .. import __version__
from ..toolkit.errors import UserError

logger = logging.getLogger(__name__)


class Migrator:
    versionsdir = os.path.dirname(__file__) + "/versions"
    UNDEFINED = object()
    LAST_ALEMBIC_VERSION = "005_statements.sql"

    def __init__(self):
        self.current_version = self.UNDEFINED
        self.versions = []

    @property
    def target_version(self):
        return self.versions[-1]

    @property
    def missing_versions(self):
        try:
            i = self.versions.index(self.current_version)
        except ValueError:
            return self.versions
        else:
            return self.versions[i + 1 :]

    def apply(self, conn, version):
        path = os.path.join(self.versionsdir, version)
        with open(path) as fo:
            sql = fo.read()
        with conn:
            with conn.cursor() as cur:
                cur.execute(sql)
                cur.execute(
                    dedent("""\
                INSERT INTO application.schema_migration_log
                (version, comment)
                VALUES (%s, 'temBoard ' || %s);
                """),
                    (version, __version__),
                )

    def check(self):
        current = self.current_version
        if current is self.UNDEFINED:
            raise Exception("Unknown revision.")

        if self.missing_versions:
            logger.info("Database version is %s.", self.current_version)
            logger.info("Target version is %s.", self.target_version)
            logger.debug(
                "Database is %s version(s) behind.", len(self.missing_versions)
            )
            raise UserError(
                "Database is not up to date." " Please upgrade with temboard migratedb."
            )
        else:
            logger.info("temBoard database is up-to-date.")

    def inspect_current_version(self, conn):
        with conn.cursor() as cur:
            cur.execute(
                dedent("""\
            SELECT table_name
            FROM information_schema.tables
            WHERE 'application' = table_schema
                AND table_name IN ('alembic_version', 'schema_migration_log')
            """)
            )

            rows = cur.fetchall()

            if len(rows) > 1:
                raise Exception("Inconsistent database versionning.")
            elif not len(rows):
                logger.debug("temBoard database is uninitialized.")
                self.current_version = None
            else:
                (table_name,) = rows[0]

                if "alembic_version" == table_name:
                    logger.debug("temBoard database is managed by Alembic.")
                    self.current_version = self.LAST_ALEMBIC_VERSION
                else:
                    cur.execute(
                        dedent("""\
                    SELECT version
                    FROM application.schema_migration_log
                    ORDER BY 1 DESC
                    LIMIT 1;
                    """)
                    )
                    (self.current_version,) = cur.fetchone()
                logger.debug("temBoard database revision is %s.", self.current_version)
        conn.commit()
        return self.current_version

    def inspect_available_versions(self):
        self.versions = sorted(
            [os.path.basename(f) for f in glob(self.versionsdir + "/*.sql")]
        )



import operator

from past.utils import old_div


def bootstrap_checks(hostinfo):
    # Default checks with thresholds to run against monitoring data

    # Loadaverage (value)
    # Global to host
    yield ("load1", hostinfo["n_cpu"] / float(2), hostinfo["n_cpu"])
    # CPU (percent)
    # One per CPU
    yield ("cpu_core", 50, 80)
    # Memory usage (percent)
    # global
    yield ("memory_usage", 50, 80)
    # Swap usage (percent)
    # Global to host
    yield ("swap_usage", 30, 50)
    # Filesystems usage (percent)
    # One per filesystem
    yield ("fs_usage_mountpoint", 80, 90)
    # Number of WAL files ready to be archived (value)
    # Global to postgres instance
    yield ("wal_files_archive", 10, 20)
    # Number of WAL files (value)
    # Global to postgres instance
    yield ("wal_files_total", 50, 100)
    # Number of transaction rollback (value)
    # One per DB
    yield ("rollback_db", 10, 20)
    # Cache hitratio (percent)
    # One per DB
    yield ("hitreadratio_db", 90, 80)
    # Client sessions vs max_connections (percent)
    # Global to postgres instance
    yield ("sessions_usage", 80, 90)
    # Waiting sessions (value)
    # One per DB
    yield ("waiting_sessions_db", 5, 10)
    # Replication lag
    # warning: 1MB, critical: 10MB
    yield ("replication_lag", 1024 * 1024, 10 * 1024 * 1024)
    # Replication connection
    yield ("replication_connection", 0, 0)
    # Temporary files size delta
    # warning: 1MB, critical: 10MB
    yield ("temp_files_size_delta", 1024 * 1024, 10 * 1024 * 1024)
    # Heap bloat ratio
    yield ("heap_bloat", 30, 50)
    # Btree bloat ratio
    yield ("btree_bloat", 30, 50)


class PreProcess:
    @staticmethod
    def loadaverage(data):
        return float(data["loadavg"][0]["load1"])

    @staticmethod
    def cpu(data):
        _data = dict()
        for r in data["cpu"]:
            total = 0
            idle = 0
            total += int(r["time_system"])
            total += int(r["time_steal"])
            total += int(r["time_iowait"])
            total += int(r["time_user"])
            total += int(r["time_idle"])
            idle = int(r["time_idle"])
            _data[r["cpu"]] = int((total - idle) / float(total) * 100)
        return _data

    @staticmethod
    def memory(data):
        return int(
            (
                int(data["memory"][0]["mem_total"])
                - int(data["memory"][0]["mem_free"])
                - int(data["memory"][0]["mem_cached"])
            )
            / float(data["memory"][0]["mem_total"])
            * 100
        )

    @staticmethod
    def swap(data):
        return int(
            int(data["memory"][0]["swap_used"])
            / float(data["memory"][0]["swap_total"])
            * 100
        )

    @staticmethod
    def fs(data):
        _data = dict()
        for r in data["filesystems_size"]:
            _data[r["mount_point"]] = int(int(r["used"]) / float(r["total"]) * 100)
        return _data

    @staticmethod
    def archive_ready(data):
        return int(data["wal_files"][0]["archive_ready"])

    @staticmethod
    def wal_files(data):
        return int(data["wal_files"][0]["total"])

    @staticmethod
    def xacts_rollback(data):
        _data = dict()
        for r in data["xacts"]:
            _data[r["dbname"]] = int(r["n_rollback"])
        return _data

    @staticmethod
    def hitratio(data):
        _data = dict()
        for r in data["blocks"]:
            hit = int(r["blks_hit"])
            read = int(r["blks_read"])
            _data[r["dbname"]] = (
                (old_div(100 * hit, (hit + read))) if read + hit > 0 else 100
            )
        return _data

    @staticmethod
    def sessions(data):
        n = 0
        for r in data["sessions"]:
            n += int(r["idle_in_xact"])
            n += int(r["idle_in_xact_aborted"])
            n += int(r["no_priv"])
            n += int(r["idle"])
            n += int(r["disabled"])
            n += int(r["waiting"])
            n += int(r["active"])
            n += int(r["fastpath"])
        return int(n / float(data["max_connections"]) * 100)

    @staticmethod
    def waiting(data):
        _data = dict()
        for r in data["sessions"]:
            _data[r["dbname"]] = int(r["waiting"])
        return _data

    @staticmethod
    def replication_lag(data):
        if data["replication_lag"]:
            return int(data["replication_lag"][0]["lag"])
        raise Exception("No replication lag info for this instance")

    @staticmethod
    def replication_connection(data):
        if data["replication_connection"]:
            k = data["replication_connection"][0]["upstream"]
            v = int(data["replication_connection"][0]["connected"])
            return {k: v}
        raise Exception("No replication connection info for this instance")

    @staticmethod
    def temp_files_size_delta(data):
        _data = dict()
        for r in data["temp_files_size_delta"]:
            _data[r["dbname"]] = int(r["size"])
        return _data

    @staticmethod
    def heap_bloat(data):
        _data = dict()
        for r in data["heap_bloat"]:
            _data[r["dbname"]] = int(r["ratio"])
        return _data

    @staticmethod
    def btree_bloat(data):
        _data = dict()
        for r in data["btree_bloat"]:
            _data[r["dbname"]] = int(r["ratio"])
        return _data


check_specs = dict(
    load1=dict(
        category="system",
        description="Loadaverage",
        preprocess=PreProcess.loadaverage,
        message="{value} is greater than {threshold}",
        operator=operator.gt,
    ),
    cpu_core=dict(
        category="system",
        description="CPU usage",
        preprocess=PreProcess.cpu,
        message="{value}% is greater than {threshold}%",
        operator=operator.gt,
        value_type="percent",
    ),
    memory_usage=dict(
        category="system",
        description="Memory usage",
        preprocess=PreProcess.memory,
        message="{value}% is greater than {threshold}%",
        operator=operator.gt,
        value_type="percent",
    ),
    swap_usage=dict(
        category="system",
        description="Swap usage",
        preprocess=PreProcess.swap,
        message="{value}% is greater than {threshold}%",
        operator=operator.gt,
        value_type="percent",
    ),
    fs_usage_mountpoint=dict(
        category="system",
        description="File systems usage",
        preprocess=PreProcess.fs,
        message="{key}: {value}% is greater than {threshold}%",
        operator=operator.gt,
        value_type="percent",
    ),
    wal_files_archive=dict(
        category="postgres",
        description="WAL files ready to be archived",
        preprocess=PreProcess.archive_ready,
        message="{value} is greater than {threshold}",
        operator=operator.gt,
    ),
    wal_files_total=dict(
        category="postgres",
        description="WAL files",
        preprocess=PreProcess.wal_files,
        message="{value} is greater than {threshold}",
        operator=operator.gt,
    ),
    rollback_db=dict(
        category="postgres",
        description="Rollbacked transactions",
        preprocess=PreProcess.xacts_rollback,
        message="{key}: {value} is greater than {threshold}",
        operator=operator.gt,
    ),
    hitreadratio_db=dict(
        category="postgres",
        description="Cache Hit Ratio",
        preprocess=PreProcess.hitratio,
        message="{key}: {value} is less than {threshold}",
        operator=operator.lt,
        value_type="percent",
    ),
    sessions_usage=dict(
        category="postgres",
        description="Client sessions",
        preprocess=PreProcess.sessions,
        message="{value} is greater than {threshold}",
        operator=operator.gt,
        value_type="percent",
    ),
    waiting_sessions_db=dict(
        category="postgres",
        description="Waiting sessions",
        preprocess=PreProcess.waiting,
        message="{key}: {value} is greater than {threshold}",
        operator=operator.gt,
    ),
    replication_lag=dict(
        category="postgres",
        description="Streaming replication lag",
        preprocess=PreProcess.replication_lag,
        message="{key}: {value} is greater than {threshold}",
        operator=operator.gt,
    ),
    replication_connection=dict(
        category="postgres",
        description="Streaming replication connection",
        preprocess=PreProcess.replication_connection,
        message="{key}: not connected",
        operator=operator.eq,
    ),
    temp_files_size_delta=dict(
        category="postgres",
        description="Temporary files size (delta)",
        preprocess=PreProcess.temp_files_size_delta,
        message="{key}: {value} is greater than {threshold}",
        operator=operator.gt,
    ),
    heap_bloat=dict(
        category="postgres",
        description="Heap bloat ratio estimation",
        preprocess=PreProcess.heap_bloat,
        message="{key}: {value} is greater than {threshold}",
        operator=operator.gt,
    ),
    btree_bloat=dict(
        category="postgres",
        description="BTree index bloat ratio estimation",
        preprocess=PreProcess.btree_bloat,
        message="{key}: {value} is greater than {threshold}",
        operator=operator.gt,
    ),
)


def get_highest_state(states):
    """
    Returns the highest state.
    """
    levels = ["UNDEF", "OK", "WARNING", "CRITICAL"]
    return levels[max([levels.index(state) for state in states])]


def checks_info(session, host_id, instance_id):
    """
    Returns alerting checks with current state by host_id/instance_id
    """
    query = """
SELECT c.name, c.warning, c.critical, c.description, c.enabled,
json_agg(json_build_object('key', cs.key, 'state', cs.state)) AS state_by_key
FROM monitoring.checks c JOIN monitoring.check_states cs ON (c.check_id = cs.check_id)
WHERE host_id = :host_id AND instance_id = :instance_id
GROUP BY 1,2,3,4,5 ORDER BY 1
    """  # noqa
    res = session.execute(query, dict(host_id=host_id, instance_id=instance_id))
    ret = []
    for row in res.fetchall():
        c_row = dict(row)
        c_row["state"] = get_highest_state([o["state"] for o in c_row["state_by_key"]])
        ret.append(c_row)
    return ret


def check_state_detail(session, host_id, instance_id, check_name):
    query = """
SELECT json_agg(json_build_object('key', cs.key,
                                  'state', cs.state,
                                  'enabled', c.enabled,
                                  'datetime', sc.datetime,
                                  'value', sc.value,
                                  'warning', sc.warning,
                                  'critical', sc.critical)) AS state_detail
FROM monitoring.checks c JOIN monitoring.check_states cs ON (c.check_id = cs.check_id)
JOIN monitoring.state_changes sc ON (sc.check_id = c.check_id AND sc.key = cs.key
                                     AND sc.datetime = (SELECT MAX(datetime)
                                                        FROM monitoring.state_changes sc2
                                                        WHERE sc2.check_id=c.check_id
                                                        AND sc2.key = cs.key
                                                        AND sc2.state = cs.state))
WHERE host_id = :host_id AND instance_id = :instance_id AND c.name = :check_name
    """  # noqa
    res = session.execute(
        query, dict(host_id=host_id, instance_id=instance_id, check_name=check_name)
    )
    row = res.fetchone()
    c_row = dict(row)
    return c_row["state_detail"]



from flask import current_app, render_template

from ...web.flask import instance_routes

PLUGIN_NAME = "maintenance"


@instance_routes.route("/maintenance")
def maintenance():
    current_app.instance.check_active_plugin(PLUGIN_NAME)
    current_app.instance.fetch_status()
    return render_template("maintenance/index.html", plugin=PLUGIN_NAME)


@instance_routes.route("/maintenance/<database>/schema/<schema>/table/<table>")
def table(database, schema, table):
    current_app.instance.check_active_plugin(PLUGIN_NAME)
    current_app.instance.fetch_status()
    return render_template(
        "maintenance/table.html",
        plugin=PLUGIN_NAME,
        database=database,
        schema=schema,
        table=table,
    )


@instance_routes.route("/maintenance/<database>/schema/<schema>")
def schema(database, schema):
    current_app.instance.check_active_plugin(PLUGIN_NAME)
    current_app.instance.fetch_status()
    return render_template(
        "maintenance/schema.html", plugin=PLUGIN_NAME, database=database, schema=schema
    )


@instance_routes.route("/maintenance/<database>")
def database(database):
    current_app.instance.check_active_plugin(PLUGIN_NAME)
    current_app.instance.fetch_status()
    return render_template(
        "maintenance/database.html", plugin=PLUGIN_NAME, database=database
    )



import logging
from contextlib import closing

from psycopg2 import connect

from ..model import format_dsn
from ..model.migrator import Migrator
from ..toolkit.app import SubCommand
from ..toolkit.errors import UserError
from .app import app

logger = logging.getLogger(__name__)


@app.command
class MigrateDB(SubCommand):
    """Manage temBoard own database."""

    def main(self, args):
        raise UserError("Missing sub-command. See --help for details.")

    def init_migrator(self):
        migrator = Migrator()
        migrator.inspect_available_versions()
        return migrator

    def make_conn(self):
        return closing(connect(format_dsn(self.app.config.repository)))


@MigrateDB.command
class Check(SubCommand):
    """Check schema synchronisation status only."""

    def main(self, args):
        migrator = self.parent.init_migrator()

        with self.parent.make_conn() as conn:
            migrator.inspect_current_version(conn)
            migrator.check()


@MigrateDB.command
class Upgrade(SubCommand):
    """Upgrade temBoard database to latest revision."""

    def main(self, args):
        migrator = self.parent.init_migrator()

        with self.parent.make_conn() as conn:
            migrator.inspect_current_version(conn)
            for version in migrator.missing_versions:
                logger.info("Upgrading database to version %s.", version)
                migrator.apply(conn, version)

        logger.info("Database up to date.")
